#pragma once
typedef complex<double> dcomp;

//----------------------------------------------------------------
// O
//----------------------------------------------------------------

inline double outer_prod( const dcomp& a, const dcomp& b )
{
	return a.real()*b.imag() - a.imag()*b.real();
}

//----------------------------------------------------------------
// Epsilon
//----------------------------------------------------------------

static const double EPS = DBL_EPSILON;

inline bool is_zero( double a )
{
	return -EPS<=a && a<=EPS;
}
inline bool is_one( double a )
{
	return 1-EPS<=a && a<=1+EPS;
}
inline bool is_ordered( double a, double b, double c )
{
	return a-EPS<=b && b<=c+EPS;
}

//----------------------------------------------------------------
// ʐ
//----------------------------------------------------------------

inline double area_of_polygon( const vector<dcomp>& pts )
{
	if( pts.size() <= 2 )
		return 0.0;

	double acc = 0.0;
	for( size_t i=1; i!=pts.size()-1; ++i )
		acc += outer_prod( pts[i]-pts[0], pts[i+1]-pts[i] );

	return fabs(acc) / 2.0;
}

//----------------------------------------------------------------
// As ->AƁABs ->BƂ̌_
//----------------------------------------------------------------

inline dcomp calc_cross_pt( dcomp As, dcomp A, dcomp Bs, dcomp B )
{
	const dcomp S = Bs-As;
	double f = outer_prod(A, B);
	double q = outer_prod(S, A) / f;
	return Bs + q * B;
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------

inline bool LineSegLineSegIntersectCheck(
  const dcomp& As, const dcomp& Ae,
  const dcomp& Bs, const dcomp& Be
  )
{
  // O=[As] Bs-Ae-Be ̉] d1, d2 ̕ƂĎ擾
  dcomp a_Bs = Bs - As;
  dcomp a_Ae = Ae - As;
  dcomp a_Be = Be - As;
  double d1 = outer_prod( a_Bs, a_Ae );
  double d2 = outer_prod( a_Ae, a_Be );
  if( d1*d2 < 0 )
    return false;

  // O=[Bs] Ae-Be-As ̉] d3, d4 ̕ƂĎ擾
  dcomp b_Ae = Ae - Bs;
  dcomp b_Be = Be - Bs;
  dcomp b_As = As - Bs;
  double d3 = outer_prod( b_Ae, b_Be );
  double d4 = outer_prod( b_Be, b_As );
  if( d3*d4 < 0 )
    return false;

  if( d1*d3 > 0 )
    return true;
  if( is_zero(d1) || is_zero(d2) || is_zero(d3) || is_zero(d4) )
    return true; // [_܂ޏꍇ

  return false;
}

//----------------------------------------------------------------
// _ɂ邩H
//----------------------------------------------------------------

inline int IsPointInsidePolygon( const dcomp& p, const vector<dcomp>& vi )
{
  vector<dcomp> v(vi);
  for(int i=0; i!=v.size(); ++i)
    v[i] -= p;

  int count = 0;
  for(int i=0; i!=v.size(); ++i)
    {
      dcomp a = v[i];
      dcomp b = v[(i+1)%v.size()];
      //
      // a + s(b-a) = (x,0)
      // s = -a.y / b.y-a.y
      // x = a.x + s (b.x - a.x)
      //
      double fy = b.imag() - a.imag();
      if( fy == 0.0 )
        {
          if( is_zero(a.imag()) )
            if( is_ordered(a.real(),0,b.real())
             || is_ordered(b.real(),0,a.real()) )
              return 1;
        }
      else
        {
          double s = -a.imag() / fy;
          double x = a.real() + s * (b.real()-a.real());
          if( is_zero(x) ) {
            if( is_ordered(a.real(),0,b.real())
             || is_ordered(b.real(),0,a.real()) )
              return 1;
            return 0;
          }
          if( 0 < x )
            {
              // _XɂƂ́A̕ӂ̉̒_ł鎞̂ݐ
              if( 0+EPS<=s && s<=1-EPS )
                ++count;
              else if( is_zero(s) && a.imag()<b.imag() )
                ++count;
              else if( is_one(s) && b.imag()<a.imag() )
                ++count;
            }
        }
    }

  return count%2==1 ? 2 : 0;
}
