#include "StdAfx.h"
#include "Util.h"
#include "TrayIcon.h"

TrayIcon::TrayIcon( TrayIconEventHandler* evhan, const char* dummyWndName )
	: evhan( evhan )
{
	const char* className = enrollStdHiddenWndClass();
	const int          ID = 0x17871787; // Magic Number

	// CxgnhWindow쐬
	eventHandlerWnd = ::CreateWindowEx(
		WS_EX_TOOLWINDOW, className, dummyWndName, 0,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, current_module(), this );

	// ݒ
	config.cbSize           = sizeof(config);
	config.hWnd             = eventHandlerWnd;
	config.uID              = ID;
	config.uFlags           = NIF_MESSAGE | NIM_SETVERSION;
	config.uCallbackMessage = getStdNotifyMsg();
	config.uVersion         = NOTIFYICON_VERSION;
	send_ADD();
	send_SETVERSION();
}

TrayIcon::~TrayIcon()
{
	send_DELETE();
	::DestroyWindow( eventHandlerWnd );
}

bool TrayIcon::send_ADD()
	{ return to_bool(::Shell_NotifyIcon( NIM_ADD,    &config )); }
bool TrayIcon::send_MODIFY()
	{ return to_bool(::Shell_NotifyIcon( NIM_MODIFY, &config )); }
bool TrayIcon::send_DELETE()
	{ return to_bool(::Shell_NotifyIcon( NIM_DELETE, &config )); }
bool TrayIcon::send_SETVERSION()
	{ return to_bool(::Shell_NotifyIcon( NIM_SETVERSION, &config )); }

void TrayIcon::readd()
{
	send_ADD();
	send_SETVERSION();
}

void TrayIcon::setIcon( HICON icn )
{
	config.uFlags |= NIF_ICON;
	config.hIcon   = icn;
	send_MODIFY();
}

void TrayIcon::setTip( string str )
{
	config.uFlags |= NIF_TIP;

	if( str.size() >= sizeof(config.szTip) )
	{
		str.resize( sizeof(config.szTip)-4 );
		str += "...";
	}
	copy( str.begin(), str.end(), &config.szTip[0] );
	config.szTip[str.size()] = '\0';
	send_MODIFY();
}

UINT TrayIcon::getTaskbarCreatedMsg()
{
	static UINT msg = ::RegisterWindowMessage( "TaskbarCreated" );
	return msg;
}

UINT TrayIcon::getStdNotifyMsg()
{
	static UINT msg = ::RegisterWindowMessage( "kilib_trayicon_stdnotifmsg" );
	return msg;
}

const char* TrayIcon::enrollStdHiddenWndClass()
{
	static WNDCLASSEX wc;
	if( wc.lpszClassName == NULL )
	{
		wc.cbSize        = sizeof(wc);
		wc.lpszClassName = "kilib_trayicon_stdhiddenwnd";
		wc.hInstance     = current_module();
		wc.lpfnWndProc   = &HiddenWndProc;
		::RegisterClassEx( &wc );
	}
	return wc.lpszClassName;
}

LRESULT CALLBACK TrayIcon::HiddenWndProc( HWND wnd, UINT msg, WPARAM wp, LPARAM lp )
{
	if( msg == WM_CREATE )
	{
		::SetWindowLongPtr( wnd, GWLP_USERDATA,
		                    reinterpret_cast<LONG_PTR>(reinterpret_cast<CREATESTRUCT*>(lp)->lpCreateParams) );
	}
	else if( msg == getTaskbarCreatedMsg() )
	{
		TrayIcon* pThis =
			reinterpret_cast<TrayIcon*>( ::GetWindowLongPtr( wnd, GWLP_USERDATA ) );
		pThis->readd();
	}
	else if( msg == getStdNotifyMsg() )
	{
		TrayIcon* pThis =
			reinterpret_cast<TrayIcon*>( ::GetWindowLongPtr( wnd, GWLP_USERDATA ) );

		switch( lp )
		{
		case NIN_KEYSELECT :
		case NIN_SELECT    :
		case WM_CONTEXTMENU:
			if( HMENU menu = pThis->evhan->on_menu() )
			{
				POINT pt;
				::GetCursorPos( &pt );
				::SetForegroundWindow( wnd );
				UINT id =
					::TrackPopupMenu(
						menu, TPM_LEFTALIGN | TPM_RIGHTBUTTON | TPM_RETURNCMD | TPM_RETURNCMD,
						pt.x, pt.y, 0, wnd, NULL );
				::PostMessage( wnd, WM_NULL, 0, 0 );
				if( id )
					pThis->evhan->on_command( menu, id );
			}
			break;
		}
	}
	return ::DefWindowProc(wnd,msg,wp,lp);
}
